from typing_extensions import TypedDict

from .stats import Stats as Stats
from .users import Users as Users


class WakaTime(TypedDict):
    user: Users
    stats: Stats
    all_time_since_today: str



from typing import Literal, TypeAlias

import httpx
from httpx import Response
from sqlalchemy import select
from nonebot_plugin_orm import get_session

from ..models import User
from ..config import config
from ..shema import Stats, Users
from ..exception import UserUnboundException

api_url = config.api_url
TimeScope: TypeAlias = Literal[
    "last_7_days", "last_30_days", "last_6_months", "last_year", "all_time"
]


class API:

    _access_token_cache: dict[str, str] = {}

    @classmethod
    async def get_access_token(cls, user_id: str) -> str:
        """Get the access token from database"""
        if user_id in cls._access_token_cache:
            return cls._access_token_cache[user_id]

        session = get_session()
        async with session.begin():
            stmt = select(User).where(User.user_id == user_id)
            user = (await session.execute(stmt)).scalar()

            if not user:
                raise UserUnboundException

            cls._access_token_cache[user_id] = user.access_token
            return user.access_token

    @classmethod
    async def bind_user(cls, code: str) -> Response:
        """Get the secret access token.

        Args:
            code: The code parameter in the callback address
        """
        async with httpx.AsyncClient() as client:
            response = await client.post(
                "https://wakatime.com/oauth/token",
                data={
                    "client_id": config.client_id,
                    "client_secret": config.client_secret,
                    "redirect_uri": "https://wakatime.com/dashboard",
                    "grant_type": "authorization_code",
                    "code": code,
                },
            )
        return response

    @classmethod
    async def get_user_info(cls, user_id: str) -> Users:
        """Get user's information"""
        access_token = await cls.get_access_token(user_id)
        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"{api_url}/users/current",
                headers={"Authorization": f"Bearer {access_token}"},
            )
        return Users(**(response.json()["data"]))

    @classmethod
    async def get_user_stats(
        cls, user_id: str, scope: TimeScope = "last_7_days"
    ) -> Stats:
        """Get user's coding activity.

        Args:
            user_id: user id
            scope: "last_7_days" "last_30_days" "last_6_months" "last_year" "all_time"
        """
        assess_token = await cls.get_access_token(user_id)
        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"{api_url}/users/current/stats/{scope}",
                headers={"Authorization": f"Bearer {assess_token}"},
            )
        return Stats(**(response.json()["data"]))

    @classmethod
    async def get_all_time_since_today(cls, user_id: str) -> str:
        """Get the total time logged since account created."""
        assess_token = await cls.get_access_token(user_id)
        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"{api_url}/users/current/all_time_since_today",
                headers={"Authorization": f"Bearer {assess_token}"},
            )
        return response.json()["data"]["text"]



import re
import base64
from pathlib import Path
from typing import Literal

import httpx


def image_to_base64(image_path: Path) -> str:
    with open(image_path, "rb") as image_file:
        base64_encoded_data = base64.b64encode(image_file.read())
        base64_message = base64_encoded_data.decode("utf-8")
    return "data:image/png;base64," + base64_message


async def get_lolicon_image() -> str:
    async with httpx.AsyncClient() as client:
        response = await client.get("https://api.lolicon.app/setu/v2")
    return response.json()["data"][0]["urls"]["original"]


def parse_time(work_time: str):

    patterns = {
        "hrs": r"(\d+)\s*hrs?",
        "mins": r"(\d+)\s*mins?",
        "secs": r"(\d+)\s*secs?",
    }

    hours = minutes = seconds = 0

    for key, pattern in patterns.items():
        match = re.search(pattern, work_time)
        if match:
            if key == "hrs":
                hours = int(match.group(1))
            elif key == "mins":
                minutes = int(match.group(1))
            elif key == "secs":
                seconds = int(match.group(1))

    total_seconds = hours * 3600 + minutes * 60 + seconds
    return total_seconds


def calc_work_time_percentage(
    work_time: str, *, duration: Literal["day", "week", "month"] = "week"
):

    if duration == "day":
        total_minutes = 24 * 60
    elif duration == "week":
        total_minutes = 7 * 24 * 60
    else:
        total_minutes = 30 * 24 * 60

    total_work_seconds = parse_time(work_time)
    total_work_minutes = total_work_seconds / 60

    percentage = (total_work_minutes / total_minutes) * 100

    return percentage



"""first revision

迁移 ID: a27298ba954d
父迁移:
创建时间: 2024-07-28 15:04:34.928642

"""

from __future__ import annotations

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

revision: str = "a27298ba954d"
down_revision: str | Sequence[str] | None = None
branch_labels: str | Sequence[str] | None = ("nonebot_plugin_wakatime",)
depends_on: str | Sequence[str] | None = None


def upgrade(name: str = "") -> None:
    if name:
        return
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "wakatime",
        sa.Column("user_id", sa.String(), nullable=False),
        sa.Column("platform", sa.String(), nullable=False),
        sa.Column("access_token", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("user_id", name=op.f("pk_wakatime")),
        info={"bind_key": "nonebot_plugin_wakatime"},
    )
    # ### end Alembic commands ###


def downgrade(name: str = "") -> None:
    if name:
        return
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("wakatime")
    # ### end Alembic commands ###



from datetime import datetime

from nonebot_plugin_htmlrender import template_to_pic

from .shema import WakaTime
from .config import RESOURCES_DIR, TEMPLATES_DIR, config
from .utils import image_to_base64, get_lolicon_image, calc_work_time_percentage


async def render(data: WakaTime) -> bytes:

    data["user"]["created_at"] = datetime.strptime(
        data["user"]["created_at"], "%Y-%m-%dT%H:%M:%SZ"
    ).strftime("%b %d %Y")

    default_background = RESOURCES_DIR / "images" / "background.png"

    if config.background_source == "default":
        background_image = image_to_base64(default_background)
    elif config.background_source == "LoliAPI":
        background_image = "https://www.loliapi.com/acg/pe/"
    else:
        background_image = await get_lolicon_image()

    return await template_to_pic(
        template_path=str(TEMPLATES_DIR),
        template_name="profile.html",
        templates={
            "user": data["user"],
            "background_image": background_image,
            "insights": {
                "data": data["stats"],
                "last_week": calc_work_time_percentage(
                    data["stats"]["human_readable_total"]
                ),
                "daily_average": calc_work_time_percentage(
                    data["stats"]["human_readable_daily_average"], duration="day"
                ),
            },
            "editors": data["stats"]["editors"],
            "languages": data["stats"]["languages"],
            "all_time_since_today": data["all_time_since_today"],
        },
        pages={
            "viewport": {"width": 550, "height": 800},
            "base_url": f"file://{TEMPLATES_DIR}",
        },
    )



from nonebot_plugin_orm import Model
from sqlalchemy.orm import Mapped, mapped_column


class User(Model):

    __tablename__ = "wakatime"

    user_id: Mapped[str] = mapped_column(primary_key=True)
    """User ID"""
    platform: Mapped[str]
    """User Account Platform"""
    access_token: Mapped[str]
    """Wakatime Access Token"""



from typing_extensions import TypedDict


class Categories(TypedDict):

    name: str
    total_seconds: float
    percent: float
    digital: str
    decimal: str
    text: str
    hours: int
    minutes: int


class Project(TypedDict):

    name: str
    total_seconds: float
    percent: float
    digital: str
    decimal: str
    text: str
    hours: int
    minutes: int


class Languages(TypedDict):
    name: str
    total_seconds: float
    percent: float
    digital: str
    decimal: str
    text: str
    hours: int
    minutes: int
    seconds: int | None


class Editors(TypedDict):
    name: str
    total_seconds: float
    percent: float
    digital: str
    decimal: str
    text: str
    hours: int
    minutes: int
    seconds: int | None


class Stats(TypedDict):

    human_readable_total: str
    human_readable_total_including_other_language: str
    daily_average: float
    daily_average_including_other_language: float
    human_readable_daily_average: str
    human_readable_daily_average_including_other_language: str
    categories: list[Categories] | None
    projects: list[Project] | None
    languages: list[Languages] | None
    editors: list[Editors] | None
    user_id: str
    username: str



import asyncio
from urllib.parse import parse_qs

from nonebot import require
from nonebot.rule import Rule
from nonebot.log import logger
from httpx import ConnectTimeout
from nonebot.internal.adapter import Event
from nonebot.plugin import PluginMetadata, inherit_supported_adapters

require("nonebot_plugin_orm")
require("nonebot_plugin_alconna")
require("nonebot_plugin_htmlrender")
from nonebot_plugin_orm import async_scoped_session
from nonebot_plugin_alconna.uniseg import At, Button, UniMessage, FallbackStrategy
from nonebot_plugin_alconna import Args, Match, Option, Alconna, MsgTarget, on_alconna

from .apis import API
from . import migrations
from .models import User
from .config import Config
from .shema import WakaTime
from .render_pic import render
from .exception import UserUnboundException
from .config import config as wakatime_config

__plugin_meta__ = PluginMetadata(
    name="谁是卷王",
    description="将代码统计嵌入 Bot 中",
    usage="/wakatime",
    type="application",
    config=Config,
    homepage="https://github.com/KomoriDev/nonebot-plugin-wakatime",
    supported_adapters=inherit_supported_adapters("nonebot_plugin_alconna"),
    extra={
        "unique_name": "WakaTime",
        "orm_version_location": migrations,
        "author": "Komorebi <mute231010@gmail.com>",
        "version": "0.1.4",
    },
)

client_id = wakatime_config.client_id
client_secret = wakatime_config.client_secret
redirect_uri = wakatime_config.redirect_uri

if client_id == "" or client_secret == "" or redirect_uri == "":
    logger.warning("缺失必要配置项，已禁用该插件")


def is_enable() -> Rule:

    def _rule() -> bool:
        return client_id != "" and client_secret != "" and redirect_uri != ""

    return Rule(_rule)


wakatime = on_alconna(
    Alconna(
        "wakatime",
        Args["target?", At | int],
        Option("-b|--bind|bind", Args["code?", str], help_text="绑定 wakatime"),
    ),
    aliases={"waka"},
    rule=is_enable(),
    use_cmd_start=True,
)


@wakatime.assign("$main")
async def _(event: Event, target: Match[At | int]):
    if target.available:
        if isinstance(target, At):
            target_name = "他"
            target_id = target.target
        else:
            target_name = "他"
            target_id = target.result
    else:
        target_name = "你"
        target_id = event.get_user_id()

    try:
        user_info_task = API.get_user_info(target_id)
        stats_info_task = API.get_user_stats(target_id)
        all_time_since_today_task = API.get_all_time_since_today(target_id)

        user_info, stats_info, all_time_since_today = await asyncio.gather(
            user_info_task, stats_info_task, all_time_since_today_task
        )
    except UserUnboundException:
        await UniMessage.text(
            f"{target_name}还没有绑定 Wakatime 账号！请私聊我并使用 /bind 命令进行绑定"
        ).finish(at_sender=True)
    except ConnectTimeout:
        await (
            UniMessage.text("网络超时，再试试叭")
            .keyboard(Button("input", "重试", text="/wakatime"))
            .finish(at_sender=True, fallback=FallbackStrategy.ignore)
        )

    result = WakaTime(
        user=user_info, stats=stats_info, all_time_since_today=all_time_since_today
    )
    image = await render(result)
    await UniMessage.image(raw=image).finish(at_sender=True)


@wakatime.assign("bind")
async def _(
    code: Match[str], event: Event, msg_target: MsgTarget, session: async_scoped_session
):

    if await session.get(User, event.get_user_id()):
        await UniMessage("已绑定过 wakatime 账号").finish(at_sender=True)

    if not msg_target.private:
        await UniMessage("绑定指令只允许在私聊中使用").finish(at_sender=True)

    if not code.available:
        auth_url = (
            f"https://wakatime.com/oauth/authorize?client_id={client_id}&response_type=code"
            f"&redirect_uri={redirect_uri}&scope=read_stats"
        )

        await (
            UniMessage.text(f"前往该页面绑定 wakatime 账号：{auth_url}")
            .keyboard(Button("link", label="即刻前往", url=auth_url))
            .finish(at_sender=True, fallback=FallbackStrategy.ignore)
        )

    resp = await API.bind_user(code.result)

    if resp.status_code == 200:
        parsed_data = parse_qs(resp.text)
        user = User(
            user_id=event.get_user_id(),
            platform=msg_target.adapter,
            access_token=parsed_data["access_token"][0],
        )
        session.add(user)
        await session.commit()
        await UniMessage("绑定成功").finish(at_sender=True)

    logger.error(f"用户 {event.get_user_id()} 绑定失败。状态码：{resp.status_code}")
    await UniMessage("绑定失败").finish(at_sender=True)



from pathlib import Path
from typing import Literal

from pydantic import Field, BaseModel
from nonebot.plugin import get_plugin_config

RESOURCES_DIR: Path = Path(__file__).parent / "resources"
TEMPLATES_DIR: Path = RESOURCES_DIR / "templates"


class ScopedConfig(BaseModel):

    client_id: str = ""
    """Your App ID from https://wakatime.com/apps"""
    client_secret: str = ""
    """Your App Secret from https://wakatime.com/apps"""
    redirect_uri: str = ""
    """Authorized Redirect URI in https://wakatime.com/apps"""
    api_url: str = "https://wakatime.com/api/v1"
    """wakatime api"""
    background_source: Literal["default", "LoliAPI", "Lolicon"] = "default"
    """Background Source"""


class Config(BaseModel):

    wakatime: ScopedConfig = Field(default_factory=ScopedConfig)
    """Wakatime Plugin Config"""


config = get_plugin_config(Config).wakatime
